@@ -1,4 +1,5 @@
 {-# LANGUAGE CPP #-}
+{-# LANGUAGE RankNTypes #-}
 {-# LANGUAGE StandaloneDeriving #-}
 {-# LANGUAGE TypeSynonymInstances #-}
 {-# LANGUAGE TemplateHaskell #-}
@@ -69,6 +70,10 @@ import Control.Monad.Writer (WriterT(WriterT), runWriterT)
 import qualified Control.Monad.Trans as MTL (lift)
 import Instances.TH.Lift ()
 
+#if !(MIN_VERSION_template_haskell(2,8,0))
+import Unsafe.Coerce (unsafeCoerce)
+#endif
+
 -- Thanks to Richard Eisenberg, GHC 7.10 adds many of the instances
 -- from this module.
 #if !MIN_VERSION_template_haskell(2,10,0)
@@ -108,9 +113,36 @@ import qualified Generics.Deriving.TH as Generic (deriveAll)
 # endif
 #endif
 
+#if !(MIN_VERSION_template_haskell(2,11,0))
+import qualified Control.Monad.Fail as Fail
+#endif
+
+-- TODO: Once GHC 8.10 is released, this should be updated to use the
+-- proper template haskell version.  Other related usages of this #if
+-- should be replaced as well (and do not have a TODO like this).
+--
+-- #if MIN_VERSION_template_haskell(2,16,0)
+#if __GLASGOW_HASKELL__ >= 809
+import GHC.Ptr (Ptr(Ptr))
+import GHC.ForeignPtr (newForeignPtr_)
+import System.IO.Unsafe (unsafePerformIO)
+#endif
+
 #if !MIN_VERSION_template_haskell(2,11,0)
 deriving instance Show NameFlavour
 deriving instance Show NameSpace
+
+instance Fail.MonadFail Q where
+  fail s = report True s >> q (fail "Q monad failure")
+    where
+      q :: (forall m. Quasi m => m a) -> Q a
+# if MIN_VERSION_template_haskell(2,8,0)
+      q = Q
+# else
+      -- Early versions of template-haskell did not expose Q's newtype
+      -- constructor. Desperate times call for desperate measures.
+      q = unsafeCoerce
+# endif
 #endif
 
 -- Ideally, it'd be possible to use reifyManyWithoutInstances for
@@ -517,5 +549,18 @@ deriving instance Typeable Ppr
 deriving instance Typeable Quasi
 #endif
 
+#if __GLASGOW_HASKELL__ >= 809
+instance Lift Bytes where
+  lift bytes =
+    [| Bytes
+      { bytesPtr = unsafePerformIO $ newForeignPtr_ (Ptr $(litE (BytesPrimL bytes)))
+      , bytesOffset = 0
+      , bytesSize = size
+      }
+    |]
+    where
+      size = bytesSize bytes
+#endif
+
 $(reifyManyWithoutInstances ''Lift [''Info, ''Loc] (const True) >>=
   deriveLiftMany)
